import React, { useLayoutEffect, useReducer, useRef, useContext, createElement, createContext, useState, useCallback } from 'react';

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function ownKeys(object, enumerableOnly) {
  var keys = Object.keys(object);

  if (Object.getOwnPropertySymbols) {
    var symbols = Object.getOwnPropertySymbols(object);
    if (enumerableOnly) symbols = symbols.filter(function (sym) {
      return Object.getOwnPropertyDescriptor(object, sym).enumerable;
    });
    keys.push.apply(keys, symbols);
  }

  return keys;
}

function _objectSpread2(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};

    if (i % 2) {
      ownKeys(Object(source), true).forEach(function (key) {
        _defineProperty(target, key, source[key]);
      });
    } else if (Object.getOwnPropertyDescriptors) {
      Object.defineProperties(target, Object.getOwnPropertyDescriptors(source));
    } else {
      ownKeys(Object(source)).forEach(function (key) {
        Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key));
      });
    }
  }

  return target;
}

function _slicedToArray(arr, i) {
  return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _unsupportedIterableToArray(arr, i) || _nonIterableRest();
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _iterableToArrayLimit(arr, i) {
  if (typeof Symbol === "undefined" || !(Symbol.iterator in Object(arr))) return;
  var _arr = [];
  var _n = true;
  var _d = false;
  var _e = undefined;

  try {
    for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(o);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
}

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;

  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

  return arr2;
}

function _nonIterableRest() {
  throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
}

var isBrowser = typeof window !== 'undefined' && typeof window.document !== 'undefined' && typeof window.document.createElement !== 'undefined' || process.env.NODE_ENV !== 'production' && process.env.TEST_MODE === 'Browser';

var useIsomorphicLayoutEffect = isBrowser ? useLayoutEffect : function (fn) {
  return fn();
};

var error = function error(message) {
  if (process.env.NODE_ENV === 'production') {
    return;
  }
  /* eslint no-console: 0 */


  console.error(message);
};
var log = {
  error: error
};

// https://github.com/reduxjs/react-redux (MIT LICENSE)
// https://github.com/dai-shi/use-context-selector (MIT LICENSE)

var createUseSelector = function createUseSelector(getContextValue, subscription, getState) {
  var defaultSelector = function defaultSelector(state) {
    return state;
  };

  function useSelector(selector) {
    var state = getContextValue();
    var storeSelector = selector || defaultSelector;

    var _useReducer = useReducer(function (s) {
      return s + 1;
    }, 0),
        _useReducer2 = _slicedToArray(_useReducer, 2),
        forceRender = _useReducer2[1];

    var selectedState;
    var latestSelector = useRef();
    var latestStoreState = useRef();
    var latestSelectedState = useRef();
    selectedState = latestSelectedState.current;

    try {
      if (storeSelector !== latestSelector.current || state !== latestStoreState.current) {
        selectedState = storeSelector(state);
      }
    } catch (err) {
      log.error(err);
    }

    useIsomorphicLayoutEffect(function () {
      latestSelector.current = storeSelector;
      latestStoreState.current = state;
      latestSelectedState.current = selectedState;
    });
    useIsomorphicLayoutEffect(function () {
      var refresh = function refresh() {
        var nextStore = getState();

        if (latestSelector.current === undefined || latestSelectedState.current === undefined) {
          return;
        }

        try {
          var newSelectedState = latestSelector.current(nextStore);

          if (newSelectedState === latestSelectedState.current) {
            return;
          }

          latestSelectedState.current = newSelectedState;
        } catch (err) {
          log.error(err);
        }

        forceRender();
      };

      subscription.add(refresh);
      return function () {
        subscription["delete"](refresh);
      };
    }, [getContextValue, subscription]);
    return selectedState;
  }

  return useSelector;
};

var entries = function entries(values) {
  return Object.entries(values);
};

var createStore = function createStore(baseContext, getCurrentState) {
  var store = {};
  entries(baseContext).forEach(function (_ref) {
    var _ref2 = _slicedToArray(_ref, 2),
        displayName = _ref2[0],
        _ref2$ = _ref2[1],
        state = _ref2$.state,
        _dispatch = _ref2$.dispatch,
        subscription = _ref2$.subscription;

    var getStateContextValue = function getStateContextValue() {
      var getStateValue = useContext(state);
      return getStateValue();
    };

    var getCurrentStateValue = function getCurrentStateValue() {
      var currentState = getCurrentState();
      return currentState[displayName];
    };

    store[displayName] = {
      state: createUseSelector(getStateContextValue, subscription, getCurrentStateValue),
      dispatch: function dispatch() {
        return useContext(_dispatch);
      }
    };
  });
  return store;
};

var dispatchEventLister = function dispatchEventLister(provider, subscription, displayName) {
  var dispatcher = /*#__PURE__*/React.memo(function (_ref) {
    var value = _ref.value,
        children = _ref.children;
    useIsomorphicLayoutEffect(function () {
      subscription.forEach(function (listener) {
        listener();
      });
    });
    return /*#__PURE__*/createElement(provider, {
      value: value
    }, children);
  }); // A bug where the displayname is anonymous will be fixed in react v17
  // see: https://github.com/facebook/react/issues/18026#issuecomment-675900452

  dispatcher.displayName = displayName;
  return dispatcher;
};

var createBaseContext = function createBaseContext(contextSource) {
  var baseContext = {};
  entries(contextSource).forEach(function (_ref2) {
    var _ref3 = _slicedToArray(_ref2, 1),
        displayName = _ref3[0];

    var stateContext = /*#__PURE__*/createContext(null, function () {
      return 0;
    });
    var dispatchContext = /*#__PURE__*/createContext(null);
    var subscription = new Set();
    stateContext.Provider = dispatchEventLister(stateContext.Provider, subscription, displayName);
    stateContext.displayName = displayName;
    dispatchContext.displayName = displayName;
    baseContext[displayName] = {
      state: stateContext,
      dispatch: dispatchContext,
      subscription: subscription
    };
  });
  return baseContext;
};

var createCurrentState = function createCurrentState(initialState) {
  var currentState = _objectSpread2({}, initialState);

  var getCurrentState = function getCurrentState() {
    return currentState;
  };

  var setCurrentState = function setCurrentState(value, key) {
    currentState[key] = value;
  };

  var resetCurrentState = function resetCurrentState() {
    currentState = _objectSpread2({}, initialState);
  };

  return {
    getCurrentState: getCurrentState,
    setCurrentState: setCurrentState,
    resetCurrentState: resetCurrentState
  };
};

var isFunction = function isFunction(value) {
  return value && {}.toString.call(value) === '[object Function]';
};

var createUseServerSideDispatch = function createUseServerSideDispatch(getCurrentState, setCurrentState, displayName, subscription) {
  /* eslint no-param-reassign: 0 */
  function useServerSideDispatch(state) {
    var currentState = getCurrentState()[displayName];
    var newState;

    if (isFunction(state)) {
      newState = state(currentState);
    } else {
      newState = state;
    }

    setCurrentState(newState, displayName);
    subscription.forEach(function (listener) {
      listener();
    });
  }

  return useServerSideDispatch;
};

var createUseStateContext = function createUseStateContext(contextSource) {
  var _createCurrentState = createCurrentState(contextSource),
      getCurrentState = _createCurrentState.getCurrentState,
      setCurrentState = _createCurrentState.setCurrentState;

  var context = createBaseContext(contextSource);
  var store = createStore(context, getCurrentState);

  var contextProvider = function contextProvider(_ref) {
    var children = _ref.children,
        value = _ref.value;
    return /*#__PURE__*/React.createElement(React.Fragment, null, entries(context).reduceRight(function (acc, _ref2) {
      var _ref3 = _slicedToArray(_ref2, 2),
          displayName = _ref3[0],
          _ref3$ = _ref3[1],
          State = _ref3$.state,
          Dispatch = _ref3$.dispatch;

      var initialValue = value && value[displayName] !== undefined ? value[displayName] : contextSource[displayName];

      var _useState = useState(initialValue),
          _useState2 = _slicedToArray(_useState, 2),
          state = _useState2[0],
          dispatch = _useState2[1];

      var getState = useCallback(function () {
        return state;
      }, [state]);
      setCurrentState(state, displayName);
      return /*#__PURE__*/React.createElement(State.Provider, {
        value: getState
      }, /*#__PURE__*/React.createElement(Dispatch.Provider, {
        value: dispatch
      }, acc));
    }, children));
  };

  var contextServerSideProvider = function contextServerSideProvider(_ref4) {
    var children = _ref4.children,
        value = _ref4.value;
    return /*#__PURE__*/React.createElement(React.Fragment, null, entries(context).reduceRight(function (acc, _ref5) {
      var _ref6 = _slicedToArray(_ref5, 2),
          displayName = _ref6[0],
          _ref6$ = _ref6[1],
          State = _ref6$.state,
          Dispatch = _ref6$.dispatch,
          subscription = _ref6$.subscription;

      var initialValue = value && value[displayName] !== undefined ? value[displayName] : contextSource[displayName];
      useIsomorphicLayoutEffect(function () {
        setCurrentState(initialValue, displayName);
      }, []);
      var dispatch = createUseServerSideDispatch(getCurrentState, setCurrentState, displayName, subscription);
      var getState = useCallback(function () {
        var currentState = getCurrentState();
        return currentState[displayName];
      }, [getCurrentState, displayName]);
      return /*#__PURE__*/React.createElement(State.Provider, {
        value: getState
      }, /*#__PURE__*/React.createElement(Dispatch.Provider, {
        value: dispatch
      }, acc));
    }, children));
  };

  return {
    context: context,
    store: store,
    contextProvider: isBrowser ? contextProvider : contextServerSideProvider,
    getState: getCurrentState
  };
};

/**
 * *useState* to create multiple contexts.
 * The created contexts are split into a state and a dispatch,
 * respectively, to prevent unnecessary rendering.
 */
var createUseStateContexts = function createUseStateContexts(contextSource) {
  var _createUseStateContex = createUseStateContext(contextSource),
      store = _createUseStateContex.store,
      contextProvider = _createUseStateContex.contextProvider,
      getState = _createUseStateContex.getState;

  return [store, contextProvider, getState];
};

var createStore$1 = function createStore(baseContext, getCurrentState) {
  var store = {};
  entries(baseContext).forEach(function (_ref) {
    var _ref2 = _slicedToArray(_ref, 2),
        displayName = _ref2[0],
        _ref2$ = _ref2[1],
        state = _ref2$.state,
        _dispatch = _ref2$.dispatch,
        subscription = _ref2$.subscription;

    var getStateContextValue = function getStateContextValue() {
      var getStateValue = useContext(state);
      return getStateValue();
    };

    var getCurrentStateValue = function getCurrentStateValue() {
      var currentState = getCurrentState();
      return currentState[displayName];
    };

    store[displayName] = {
      state: createUseSelector(getStateContextValue, subscription, getCurrentStateValue),
      dispatch: function dispatch() {
        return useContext(_dispatch);
      }
    };
  });
  return store;
};

var getInitialState = function getInitialState(contextSource) {
  return entries(contextSource).reduce(function (acc, _ref) {
    var _ref2 = _slicedToArray(_ref, 2),
        displayName = _ref2[0],
        initialState = _ref2[1].initialState;

    acc[displayName] = initialState;
    return acc;
  }, {});
};

var createUseServerSideDispatch$1 = function createUseServerSideDispatch(getCurrentState, setCurrentState, displayName, reducer, subscription) {
  var useServerSideDispatch = function useServerSideDispatch(action) {
    /* eslint no-param-reassign: 0 */
    var currentState = getCurrentState()[displayName];

    if (reducer.length === 1) {
      setCurrentState(reducer(currentState, undefined), displayName);
    } else {
      setCurrentState(reducer(currentState, action), displayName);
    }

    subscription.forEach(function (listener) {
      listener();
    });
  };

  return useServerSideDispatch;
};

var createUseReducerContext = function createUseReducerContext(contextSource) {
  var _createCurrentState = createCurrentState(getInitialState(contextSource)),
      getCurrentState = _createCurrentState.getCurrentState,
      setCurrentState = _createCurrentState.setCurrentState;

  var context = createBaseContext(contextSource);
  var store = createStore$1(context, getCurrentState);

  var contextProvider = function contextProvider(_ref3) {
    var children = _ref3.children,
        value = _ref3.value;
    return /*#__PURE__*/React.createElement(React.Fragment, null, entries(context).reduceRight(function (acc, _ref4) {
      var _ref5 = _slicedToArray(_ref4, 2),
          displayName = _ref5[0],
          _ref5$ = _ref5[1],
          State = _ref5$.state,
          Dispatch = _ref5$.dispatch;

      var _contextSource$displa = contextSource[displayName],
          reducer = _contextSource$displa.reducer,
          initialState = _contextSource$displa.initialState,
          initializer = _contextSource$displa.initializer;
      var initialValue = value && value[displayName] !== undefined ? value[displayName] : initialState;

      var _useReducer = useReducer(reducer, initialValue, initializer),
          _useReducer2 = _slicedToArray(_useReducer, 2),
          state = _useReducer2[0],
          dispatch = _useReducer2[1];

      var getState = useCallback(function () {
        return state;
      }, [state]);
      setCurrentState(state, displayName);
      return /*#__PURE__*/React.createElement(State.Provider, {
        value: getState
      }, /*#__PURE__*/React.createElement(Dispatch.Provider, {
        value: dispatch
      }, acc));
    }, children));
  };

  var contextServerSideProvider = function contextServerSideProvider(_ref6) {
    var children = _ref6.children,
        value = _ref6.value;
    return /*#__PURE__*/React.createElement(React.Fragment, null, entries(context).reduceRight(function (acc, _ref7) {
      var _ref8 = _slicedToArray(_ref7, 2),
          displayName = _ref8[0],
          _ref8$ = _ref8[1],
          State = _ref8$.state,
          Dispatch = _ref8$.dispatch,
          subscription = _ref8$.subscription;

      var initialState = contextSource[displayName].initialState;
      var initialValue = value && value[displayName] !== undefined ? value[displayName] : initialState;
      useIsomorphicLayoutEffect(function () {
        setCurrentState(initialValue, displayName);
      }, []);
      var dispatch = createUseServerSideDispatch$1(getCurrentState, setCurrentState, displayName, contextSource[displayName].reducer, subscription);
      var getState = useCallback(function () {
        var currentState = getCurrentState();
        return currentState[displayName];
      }, [getCurrentState, displayName]);
      return /*#__PURE__*/React.createElement(State.Provider, {
        value: getState
      }, /*#__PURE__*/React.createElement(Dispatch.Provider, {
        value: dispatch
      }, acc));
    }, children));
  };

  return {
    store: store,
    contextProvider: isBrowser ? contextProvider : contextServerSideProvider,
    getState: getCurrentState
  };
};

/**
 * *useReducer* to create multiple contexts.
 * The created contexts are split into a state and a dispatch,
 * respectively, to prevent unnecessary rendering.
 */
var createUseReducerContexts = function createUseReducerContexts(contextSource) {
  var _createUseReducerCont = createUseReducerContext(contextSource),
      store = _createUseReducerCont.store,
      contextProvider = _createUseReducerCont.contextProvider,
      getState = _createUseReducerCont.getState;

  return [store, contextProvider, getState];
};

export { createUseReducerContexts, createUseStateContexts };
